import cv2
import math
import numpy as np


class Contours:
    @staticmethod
    def area(contour):
        """
        @brief Calculates a contour area

        The function computes a contour area. Similarly to moments , the area is computed using the Green
        formula. The function will most certainly give a wrong
        results for contours with self-intersections.

        :param contour:
        :return:
        """
        return cv2.contourArea(contour)

    @staticmethod
    def perimeter(contour):
        """
        @brief Calculates a contour perimeter

        The function computes a closed contour perimeter.

        :param contour: Input vector of 2D points, stored in std::vector or Mat
        :return:
        """
        return cv2.arcLength(contour, True)

    @staticmethod
    def convex_hull(contour):
        """
        @brief Finds the convex hull of a point set.

        The functions find the convex hull of a 2D point set using the Sklansky's algorithm that has O(N logN)
        complexity in the current implementation on OpenCV.

        :param contour: Input 2D point set, stored in std::vector or Mat
        :return set of points:
        """
        return cv2.convexHull(contour)

    @staticmethod
    def convex_hull_perimeter(contour):
        """
        @brief Finds the perimeter of the convex-hull.

        :param contour:
        :return:
        """
        return Contours.perimeter(Contours.convex_hull(contour))

    @staticmethod
    def convex_hull_area(contour):
        """
        Calculates the area of the convex hull from the contour, using the same function that the area of any object.

        :param contour:
        :return:
        """
        return cv2.contourArea(Contours.convex_hull(contour))

    @staticmethod
    def bounding_box_area(contour):
        """
        @brief Calculates the area of the bounding box of the contour.

        :param contour:
        :return:
        """
        return Contours.major_axis(contour) * Contours.minor_axis(contour)

    @staticmethod
    def rectangularity(contour):
        """
        @brief Calculates the proportion between the real area of the contour and the bounding box

        The bounding box is the minimum rectangle that

        :param contour: 1 channel image
        :return:
        """
        return Contours.area(contour) / Contours.bounding_box_area(contour)

    @staticmethod
    def max_r(contour):
        """
        @brief Calculates the radius of the enclosing circle of the contour

        :param contour:
        :return:
        """
        (_, _), radius = cv2.minEnclosingCircle(contour)
        return radius

    @staticmethod
    def min_r(contour):
        """
        @brief Calculates the radius of the maximum inscribed circle of the contour

        :param contour:
        :return:
        """
        _, _, width, height = cv2.boundingRect(contour)
        width = int(width) * 2
        height = int(height) * 2
        raw_distance = np.empty((width, height), dtype=np.float32)
        for i in range(width):
            for j in range(height):
                raw_distance[i, j] = cv2.pointPolygonTest(contour, (j, i), True)
        _, maxVal, _, _ = cv2.minMaxLoc(raw_distance)
        return abs(maxVal)

    @staticmethod
    def major_axis(contour):
        """
        @brief Calculates the major axis of the enclosing ellipse of the contour

        :param contour:
        :return:
        """
        (_, _), (_, major), _ = cv2.fitEllipse(contour)

        return major

    @staticmethod
    def minor_axis(contour):
        """
        @brief Calculates the minor axis of the ellipse from the contour

        :param contour:
        :return:
        """
        (_, _), (minor, _), _ = cv2.fitEllipse(contour)
        return minor

    @staticmethod
    def orientation(contour):
        """
        @brief Find the angle orientation of the enclosing ellipse

        :param contour:
        :return:
        """

        (_, _), (_, _), angle = cv2.fitEllipse(contour)
        return angle

    @staticmethod
    def roundness(contour):
        """
        Circularity corrected by the aspect ratio

        ref : https://progearthplanetsci.springeropen.com/articles/10.1186/s40645-015-0078-x
        :param contour: 1 channel image
        :return:
        """
        return round(4 * Contours.area(contour) / (math.pi * Contours.major_axis(contour) * Contours.major_axis(contour)), 2)

    @staticmethod
    def circularity(contour):
        """
        @brief Calc the likeliness of an object to a circle

        :param contour:
        :return:
        """
        return round(4 * math.pi * Contours.area(contour) / (Contours.perimeter(contour) * Contours.perimeter(contour)),
                     2)

    @staticmethod
    def solidity(contour):
        """
        @brief Calc the proportion between the area of the contour and the convex-hull

        :param contour:
        :return:
        """

        return round(Contours.area(contour) / Contours.convex_hull_area(contour), 2)

    @staticmethod
    def sphericity(contour):
        """
        @brief Proportion between the major diagonal and the minor diagonal

        :param contour:
        :return:
        """

        return Contours.min_r(contour) / Contours.max_r(contour)

    @staticmethod
    def aspect_ratio(contour):
        """
        @brief Proportional relationship between its width and it's height
        :param contour:
        :return:
        """

        return round(Contours.major_axis(contour) / Contours.minor_axis(contour), 2)

    @staticmethod
    def area_equivalent_diameter(contour):
        """
        @brief The diamater of the real area of the contour
        :param contour:
        :return:
        """

        return math.sqrt((4 / math.pi) * Contours.area(contour))

    @staticmethod
    def perimeter_equivalent_diameter(contour):
        """
        @brief The diameter of the real perimeter of the contour
        ;param contour:
        """

        return Contours.area(contour) / math.pi

    @staticmethod
    def equivalent_ellipse_area(contour):
        """
        @brief The area of the equivalent ellipse
        :param contour:
        :return:
        """
        return math.pi * Contours.major_axis(contour) * Contours.minor_axis(contour)

    @staticmethod
    def compactness(contour):
        """
        @brief Proportion between area and the shape of the ellipse

        :param contour:
        :return:
        """
        return math.sqrt((4 / math.pi * Contours.area(contour)) / Contours.major_axis(contour))

    @staticmethod
    def convexity(contour):
        """
        @brief Calc the convexity of the contour

        The convexity is a measure of the curvature of an object. Is calc by the relation between the perimeter of
        the convex hull and the perimeter of the object.
        :param contour:
        :return:
        """
        return Contours.convex_hull_perimeter(contour) / Contours.perimeter(contour)

    @staticmethod
    def shape(contour):
        """
        Relation between perimeter and area. Calc the elongation of an object
        :param contour:
        :return:
        """
        return math.pow(Contours.perimeter(contour), 2) / Contours.area(contour)

    @staticmethod
    def r_factor(contour):
        return Contours.convex_hull_perimeter(contour) / (Contours.major_axis(contour) * math.pi)

    @staticmethod
    def eccentricity(contour):
        """
        @brief Calc how much the conic section deviates from being circular

        For any point of a conic section, the distance between a fixed point F and a fixed straight line l is always
        equal to a positive constant, the eccentricity. Is calculed by the relation between the two diagonals of the
        elipse.

        :param contour:
        :return:
        """
        ellipse = cv2.fitEllipse(contour)
        D = math.fabs((ellipse[0][0] - ellipse[1][0]))
        d = math.fabs(ellipse[0][1] - ellipse[1][1])

        return round(D / d, 2)

    @staticmethod
    def max_feret(contour):
        """
        The maximum distance between parallel tangents to the projection area of the contour
        :param contour:
        :return:
        """
        feret, _ = Contours._max_min_feret(contour)
        return feret

    @staticmethod
    def min_feret(contour):
        """
        The minimum distance between parallel tangents to the projection area of the contour
        :param contour:
        :return:
        """
        _, feret = Contours._max_min_feret(contour)
        return feret

    @staticmethod
    def elongation(contour):
        """
        Relation between maximum and minimum feret of the contour.
        :param contour:
        :return:
        """
        return Contours.max_feret(contour) / Contours.min_feret(contour)

    @staticmethod
    def hu_moments(contour):
        return cv2.HuMoments(cv2.moments(contour)).flatten()

    @staticmethod
    def center(contour):
        """
        @brief Calculate the centroid of a contour

        The centroid of a plane figure is the arithmetic mean of all the point in the figure. For calculate it we
        use the moments of the image ( see https://en.wikipedia.org/wiki/Image_moment).

        :param contour:
        :return:
        """
        m = cv2.moments(contour)
        cx = int(m['m10'] / m['m00'])
        cy = int(m['m01'] / m['m00'])

        return (cx, cy)

    @staticmethod
    def _max_min_feret(contour):
        """
        Helper method for calculation of maximum and minimum ferets based on convex hull of the contour.
        Based on C++ code: https://www.crisluengo.net/archives/408
        :param contour:
        :return:
        """
        convex_hull_contour = Contours.convex_hull(contour)
        min_feret = 999999
        max_feret = 0
        n = len(convex_hull_contour) - 1
        p0 = n
        p = 0
        q = 1

        while Contours._triangle_area(convex_hull_contour[p][0], convex_hull_contour[Contours._next_point(p, n)][0],
                            convex_hull_contour[Contours._next_point(q, n)][0]) > \
                Contours._triangle_area(convex_hull_contour[p][0], convex_hull_contour[Contours._next_point(p, n)][0],
                                        convex_hull_contour[q][0]):
            q = Contours._next_point(q, n)

        while p != p0:
            p = Contours._next_point(p, n)
            listq = [q]
            while Contours._triangle_area(convex_hull_contour[p][0], convex_hull_contour[Contours._next_point(p, n)][0],
                                          convex_hull_contour[Contours._next_point(q, n)][0]) > \
                    Contours._triangle_area(convex_hull_contour[p][0], convex_hull_contour[Contours._next_point(p, n)][0],
                                            convex_hull_contour[q][0]):
                q = Contours._next_point(q, n)
                listq.append(q)

            if Contours._triangle_area(convex_hull_contour[p][0], convex_hull_contour[Contours._next_point(p, n)][0],
                                       convex_hull_contour[Contours._next_point(q, n)][0]) == \
                    Contours._triangle_area(convex_hull_contour[p][0], convex_hull_contour[Contours._next_point(p, n)][0],
                                                                                  convex_hull_contour[q][0]):
                listq.append(Contours._next_point(q, n))

            for i in range(len(listq)):
                q = ((listq[i] - 1) % n) + 1
                d = math.sqrt((convex_hull_contour[p][0][0] - convex_hull_contour[q][0][0]) ** 2 +
                              (convex_hull_contour[p][0][1] - convex_hull_contour[q][0][1]) ** 2)
                if d > max_feret:
                    max_feret = d

            p3 = convex_hull_contour[p][0]
            for i in range(len(listq) - 2):
                p1 = convex_hull_contour[listq[i]][0]
                p2 = convex_hull_contour[listq[i + 1]][0]
                height = Contours._triangle_height(p1, p2, p3)

                if height < min_feret:
                    min_feret = height

        return max_feret, min_feret

    @staticmethod
    def _triangle_area(p1, p2, p3):
        """
        Helper method that calculates triangle area based on triangle vertices
        :param p1: First vertex of the triangle
        :param p2: Second vertex of the triangle
        :param p3: Third vertex of the triangle
        :return:
        """
        return ((p2[0] - p1[0]) * (p3[1] - p1[1]) - (p2[1] - p1[1]) * (p3[0] - p1[0])) / 2

    @staticmethod
    def _triangle_height(p1, p2, p3):
        """
        Helper method that calculates triangle height based on triangle vertices
        :param p1: First vertex of the triangle
        :param p2: Second vertex of the triangle
        :param p3: Third vertex of the triangle
        :return:
        """
        return ((p2[0] - p1[0]) * (p3[1] - p1[1]) - (p2[1] - p1[1]) * (p3[0] - p1[0])) / \
               math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)

    @staticmethod
    def _next_point(p, n):
        """
        Helper method that calculates next antipodal point
        :param p: previous point
        :param n: total number of points in a contour
        :return:
        """
        return p % n + 1



import cv2
from copy import copy
import numpy as np
from sklearn import cluster


class Color:
    @staticmethod
    def mean_sdv_lab(img, channel=0):
        """
        Calc the mean an the standard desviation of a channel of a CIE-LAB image
        :param img: Image with three channels
        :param channel: 0 => all three, 1 => L , 2 => A, 3 => B
        :return:
        """
        img_lab = cv2.cvtColor(img, cv2.COLOR_RGB2LAB)

        return Color.mean_sdv(img_lab, channel)

    @staticmethod
    def mean_sdv_rgb(img, channel=0):
        """
        Calc the mean an the standard desviation of a channel of a RGB image
        :param img: Image with three channels
        :param channel: 0 => all three, 1 => L , 2 => A, 3 => B
        :return:
        """
        return Color.mean_sdv(img, channel)

    @staticmethod
    def mean_sdv_hsv(img, channel=0):
        """
        Calc the mean an the standard desviation of a channel of a HSV image
        :param img: Image with three channels
        :param channel: 0 => all three, 1 => L , 2 => A, 3 => B
        :return:
        """
        img_lab = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)

        return Color.mean_sdv(img_lab, channel)

    @staticmethod
    def mean_sdv(img, channel=0):
        """
        @brief Calculate the mean and the starndard desviation of an image.

        :param img:
        :param channel:
        :return:
        """
        if channel != 0:
            if channel == 1:
                chann, _, _ = cv2.split(img)
            elif channel == 2:
                _, chann, _ = cv2.split(img)
            else:
                _, _, chann = cv2.split(img)

        else:
            chann = copy(img)

        return cv2.meanStdDev(chann)

    @staticmethod
    def dominant_colors(image: np.ndarray, mask: np.ndarray, n_colors: int,
                        random_start: int = 42):
        """
        @brief Get the dominants colors of an image.

        :param image:
        :param mask:
        :param n_colors:
        :param random_start:
        :return:
        """
        hue, saturation, value = cv2.split(image)

        train = []

        for x in range(0, hue.shape[0]):
            for y in range(0, hue.shape[1]):
                if mask[x][y] == 1:
                    if train is None:
                        train = np.array([hue[x][y], saturation[x][y], value[x][y]])
                    else:
                        train.append(np.array([hue[x][y], saturation[x][y], value[x][y]]))

        kmeans = cluster.KMeans(n_clusters=n_colors, random_state=random_start)
        clusters = kmeans.fit_predict(np.array(train))

        _, importance = np.unique(clusters, return_counts=True)

        importance = importance / sum(importance)

        return zip(kmeans.cluster_centers_, importance)




from skimage.feature import greycomatrix, greycoprops
import numpy as np
from scipy import stats
import cv2


class Texture:

    @staticmethod
    def texture_features(image, distances, angles, properties):
        """
        @brief Calc the texture properties of a greyscale image

        Calculate the property passed as parameter of the function.
        ref: http://scikit-image.org/docs/0.7.0/api/skimage.feature.texture.html

        :param distances: List of pixel pair distance offsets
        :param angles: List of pixel pair angles in radians.
        :param properties: Array of Strings. The feature that you want to calculate. 'contrast',
        'dissimilarity', 'homogeneity', 'ASM', 'energy', 'correlation'.
        :param image: Greyscale image in uint8
        :return:
        """
        glcm = greycomatrix(image, distances=distances, angles=angles, symmetric=True, normed=True)

        return np.hstack([greycoprops(glcm, prop).ravel() for prop in properties])

    @staticmethod
    def skew(img):
        img_histogram = cv2.calcHist([img], [0], None, [256], [0, 256])

        return stats.skew(img_histogram)

    @staticmethod
    def kurtosis(img):
        img_histogram = cv2.calcHist([img], [0], None, [256], [0, 256])

        return stats.kurtosis(img_histogram)



# coding=utf-8
from setuptools import setup

setup(name='uib_vfeatures',
      version='0.3.1',
      description='Vision features of generalistic use',
      url='https://gitlab.com/miquelca32/features',
      author="Miquel Miró Nicolau, Bernat Galmés Rubert, Dr. Gabriel Moyà Alcover",
      author_email='miquelca32@gmail.com, bernat_galmes@hotmail.com, gabriel_moya@uib.es',
      license='MIT',
      packages=['uib_vfeatures'],
      keywords=['Features extraction', 'Machine Learning', 'Computer Vision'],
      install_requires=[
          'scikit-image',
          'opencv-python==4.1.1.26',
          'matplotlib',
          'scipy',
          'scikit-image',
          'numpy'
      ],
      zip_safe=False)



from matplotlib import pyplot as plt
from copy import copy
import cv2
from uib_vfeatures.contours import Contours
import numpy as np


class Masks:

    @staticmethod
    def solidity(mask, screen=False):
        """
        Calculates the proportion between the area of the object in the mask and the convex-hull
        :param mask: 1 channel image
        :param screen:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        if screen:
            hull = cv2.convexHull(cnt, returnPoints=False)
            defects = cv2.convexityDefects(cnt, hull)
            mask_cp = copy(mask)
            for i in range(defects.shape[0]):
                s, e, f, d = defects[i, 0]
                start = tuple(cnt[s][0])
                end = tuple(cnt[e][0])
                cv2.line(mask_cp, start, end, [100, 100, 100], 10)

            plt.imshow(mask_cp)
            plt.show()

        return Contours.solidity(cnt)

    @staticmethod
    def convex_hull_perimeter(mask):
        """
        @brief Finds the perimeter of the convex-hull.

        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.convex_hull_perimeter(cnt)

    @staticmethod
    def convex_hull_area(mask):
        """
        Calculates the area of the convex hull from the contour, using the same function that the area of any object.

        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.convex_hull_area(cnt)

    @staticmethod
    def bounding_box_area(mask, screen=False):
        """
        @brief Calculates the area of the bounding box of the contour.

        :param mask: 1 channel image
        :param screen: Boolean
        :return:
        """

        cnt = Masks.extract_contour(mask)

        if screen:
            rect = cv2.minAreaRect(cnt)
            box = cv2.boxPoints(rect)
            box = np.int0(box)
            mask_cp = copy(mask)
            cv2.drawContours(mask_cp, [box], -1, (100, 100, 100), 20)

            plt.imshow(mask_cp)
            plt.show()

        return Contours.bounding_box_area(cnt)

    @staticmethod
    def rectangularity(mask):
        """
        @brief Calculates the area of the bounding box of the contour.

        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.bounding_box_area(cnt)

    @staticmethod
    def min_r(mask):
        """
        @brief Calculates the minor radius of the ellipse from the contour
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.min_r(cnt)

    @staticmethod
    def max_r(mask):
        """
        @brief Calculates the radius of the enclosing circle of the contour
        :param mask:
        :return:
        """

        cnt = Masks.extract_contour(mask)

        return Contours.max_r(cnt)

    @staticmethod
    def feret(mask):
        """
        @brief Calculates the major diagonal of the enclosing ellipse of the contour
        :param mask:
        :return:
        """

        cnt = Masks.extract_contour(mask)

        return Contours.major_axis(cnt)

    @staticmethod
    def breadth(mask):
        """
        @brief Calculates the minor diagonal of the ellipse from the contour
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.minor_axis(cnt)

    @staticmethod
    def circularity(mask):
        """
        @brief Calc the likeliness of an object to a circle
        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.circularity(cnt)

    @staticmethod
    def roundness(mask):
        """
        Circularity corrected by the aspect ratio
        ref : https://progearthplanetsci.springeropen.com/articles/10.1186/s40645-015-0078-x
        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.roundness(cnt)

    @staticmethod
    def feret_angle(mask):
        """
        @brief Return the angle between the feret and the horizontal
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.orientation(cnt)

    @staticmethod
    def eccentricity(mask, screen=False):
        """
        @brief Calc how much the conic section deviates from being circular

        For any point of a conic section, the distance between a fixed point F and a fixed straight line l is always
        equal to a positive constant, the eccentricity. Is calculed by the relation between the two diagonals of the
        elipse.

        :param mask: 1 channel image
        :param screen:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        if screen:
            mask_cp = copy(mask)
            ellipse = cv2.fitEllipse(cnt)
            cv2.ellipse(mask_cp, ellipse, (100, 100, 100), 7)
            plt.imshow(mask_cp)
            plt.show()

        return Contours.eccentricity(cnt)

    @staticmethod
    def center(mask):
        """
        @brief Calculate the centroid of a contour

        The centroid of a plane figure is the arithmetic mean of all the point in the figure.

        :param mask:  1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.center(cnt)

    @staticmethod
    def sphericity(mask):
        """
        Proportion between the major and the minor feret

        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.sphericity(cnt)

    @staticmethod
    def aspect_ratio(mask):
        """
        @brief Proportional relationship between its width and it's height
        :param mask: 2D Vector, 1 channel image
        :return:
        """

        cnt = Masks.extract_contour(mask)

        return Contours.aspect_ratio(cnt)

    @staticmethod
    def area_equivalent_diameter(mask):
        """
        @brief The diamater of the real area of the contour

        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.area_equivalent_diameter(cnt)

    @staticmethod
    def perimeter_equivalent_diameter(mask):
        """
        @brief The diameter of the real perimeter of the contour

        :param mask: 2D Vector, 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.perimeter_equivalent_diameter(cnt)

    @staticmethod
    def equivalent_ellipse_area(mask):
        """
        The area of the equivalent ellipse
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.equivalent_ellipse_area(cnt)

    @staticmethod
    def compactness(mask):
        """
        Proportion between area and the shape of the ellipse
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.compactness(cnt)

    @staticmethod
    def area(mask):
        """
        Calc the area of the object of the mask

        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return cv2.contourArea(cnt)

    @staticmethod
    def convexity(mask):
        """
        @brief Calc the convexity of the contour

        The convexity is a measure of the curvature of an object. Is calc by the relation between the perimeter of
        the convex hull and the perimeter of the object.
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.convexity(cnt)

    @staticmethod
    def shape(mask):
        """
        Relation between perimeter and area. Calc the elongation of an object
        :param mask:
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.shape(cnt)

    @staticmethod
    def perimeter(mask):
        """
        Calc the perimeter of the object in the mask

        :param mask: 1 channel image
        :return:
        """
        cnt = Masks.extract_contour(mask)

        return Contours.perimeter(cnt)

    @staticmethod
    def extract_contour(mask):
        """
        @brief Finds contours in a binary image.

        The function retrieves contours from the binary image using the algorithm @cite Suzuki85 . The contours
        are a useful tool for shape analysis and object detection and recognition.

        @note Source image is not modified by this function.

        :param mask:
        :return:
        """
        if len(mask.shape) != 2:
            raise ValueError('Image is not a maks, multiples channels of color')

        contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

        if len(contours) > 10:
            raise ValueError('Too many contours, image is not of a only object mask')

        if len(contours) == 0:
            raise ValueError("0 contours found: the image doesn't have any object")

        return contours[0]